"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ChainConnectorStore = exports.checkDBError = exports.getDBInstance = void 0;
const lisk_sdk_1 = require("lisk-sdk");
const os = require("os");
const path_1 = require("path");
const fs_extra_1 = require("fs-extra");
const constants_1 = require("./constants");
const schemas_1 = require("./schemas");
const { Database } = lisk_sdk_1.db;
const getDBInstance = async (dataPath, dbName = 'lisk-framework-chain-connector-plugin.db') => {
    const dirPath = (0, path_1.join)(dataPath.replace('~', os.homedir()), 'plugins/data', dbName);
    await (0, fs_extra_1.ensureDir)(dirPath);
    return new Database(dirPath);
};
exports.getDBInstance = getDBInstance;
const checkDBError = (error) => {
    if (!(error instanceof lisk_sdk_1.db.NotFoundError)) {
        throw error;
    }
};
exports.checkDBError = checkDBError;
class ChainConnectorStore {
    constructor(db) {
        this._db = db;
    }
    close() {
        this._db.close();
    }
    get privateKey() {
        return this._privateKey;
    }
    async getBlockHeaders() {
        let blockHeaders = [];
        try {
            const encodedInfo = await this._db.get(constants_1.DB_KEY_BLOCK_HEADERS);
            blockHeaders = lisk_sdk_1.codec.decode(schemas_1.blockHeadersInfoSchema, encodedInfo).blockHeaders;
        }
        catch (error) {
            (0, exports.checkDBError)(error);
        }
        return blockHeaders;
    }
    async setBlockHeaders(blockHeaders) {
        const encodedInfo = lisk_sdk_1.codec.encode(schemas_1.blockHeadersInfoSchema, { blockHeaders });
        await this._db.set(constants_1.DB_KEY_BLOCK_HEADERS, encodedInfo);
    }
    async getAggregateCommits() {
        let aggregateCommits = [];
        try {
            const encodedInfo = await this._db.get(constants_1.DB_KEY_AGGREGATE_COMMITS);
            aggregateCommits = lisk_sdk_1.codec.decode(schemas_1.aggregateCommitsInfoSchema, encodedInfo).aggregateCommits;
        }
        catch (error) {
            (0, exports.checkDBError)(error);
        }
        return aggregateCommits;
    }
    async setAggregateCommits(aggregateCommits) {
        const encodedInfo = lisk_sdk_1.codec.encode(schemas_1.aggregateCommitsInfoSchema, { aggregateCommits });
        await this._db.set(constants_1.DB_KEY_AGGREGATE_COMMITS, encodedInfo);
    }
    async getValidatorsHashPreimage() {
        let validatorsHashPreimage = [];
        try {
            const encodedInfo = await this._db.get(constants_1.DB_KEY_VALIDATORS_HASH_PREIMAGE);
            validatorsHashPreimage = lisk_sdk_1.codec.decode(schemas_1.validatorsHashPreimageInfoSchema, encodedInfo).validatorsHashPreimage;
        }
        catch (error) {
            (0, exports.checkDBError)(error);
        }
        return validatorsHashPreimage;
    }
    async setValidatorsHashPreimage(validatorsHashInput) {
        const encodedInfo = lisk_sdk_1.codec.encode(schemas_1.validatorsHashPreimageInfoSchema, {
            validatorsHashPreimage: validatorsHashInput,
        });
        await this._db.set(constants_1.DB_KEY_VALIDATORS_HASH_PREIMAGE, encodedInfo);
    }
    async getCrossChainMessages() {
        let crossChainMessages = [];
        try {
            const encodedInfo = await this._db.get(constants_1.DB_KEY_CROSS_CHAIN_MESSAGES);
            crossChainMessages = lisk_sdk_1.codec.decode(schemas_1.ccmsFromEventsSchema, encodedInfo).ccmsFromEvents;
        }
        catch (error) {
            (0, exports.checkDBError)(error);
        }
        return crossChainMessages;
    }
    async setCrossChainMessages(ccms) {
        const encodedInfo = lisk_sdk_1.codec.encode(schemas_1.ccmsFromEventsSchema, { ccmsFromEvents: ccms });
        await this._db.set(constants_1.DB_KEY_CROSS_CHAIN_MESSAGES, encodedInfo);
    }
    async getLastSentCCM() {
        let lastSentCCM;
        try {
            const encodedInfo = await this._db.get(constants_1.DB_KEY_LAST_SENT_CCM);
            lastSentCCM = lisk_sdk_1.codec.decode(schemas_1.lastSentCCMWithHeight, encodedInfo);
        }
        catch (error) {
            (0, exports.checkDBError)(error);
        }
        return lastSentCCM;
    }
    async setLastSentCCM(lastSentCCM) {
        await this._db.set(constants_1.DB_KEY_LAST_SENT_CCM, lisk_sdk_1.codec.encode(schemas_1.lastSentCCMWithHeight, lastSentCCM));
    }
    async getListOfCCUs() {
        let listOfCCUs = [];
        try {
            const encodedInfo = await this._db.get(constants_1.DB_KEY_LIST_OF_CCU);
            listOfCCUs = lisk_sdk_1.codec.decode(schemas_1.listOfCCUsSchema, encodedInfo).listOfCCUs;
        }
        catch (error) {
            (0, exports.checkDBError)(error);
        }
        return listOfCCUs;
    }
    async setListOfCCUs(listOfCCUs) {
        listOfCCUs.sort((a, b) => Number(b.nonce) - Number(a.nonce));
        await this._db.set(constants_1.DB_KEY_LIST_OF_CCU, lisk_sdk_1.codec.encode(schemas_1.listOfCCUsSchema, { listOfCCUs }));
    }
    async setPrivateKey(encryptedPrivateKey, password) {
        const parsedEncryptedKey = lisk_sdk_1.cryptography.encrypt.parseEncryptedMessage(encryptedPrivateKey);
        this._privateKey = Buffer.from(await lisk_sdk_1.cryptography.encrypt.decryptMessageWithPassword(parsedEncryptedKey, password, 'utf-8'), 'hex');
    }
    async deletePrivateKey(encryptedPrivateKey, password) {
        const parsedEncryptedKey = lisk_sdk_1.cryptography.encrypt.parseEncryptedMessage(encryptedPrivateKey);
        await lisk_sdk_1.cryptography.encrypt.decryptMessageWithPassword(parsedEncryptedKey, password, 'utf-8');
        this._privateKey = undefined;
    }
}
exports.ChainConnectorStore = ChainConnectorStore;
//# sourceMappingURL=db.js.map