"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.MonitorPlugin = void 0;
const lisk_sdk_1 = require("lisk-sdk");
const express = require("express");
const cors = require("cors");
const express_rate_limit_1 = require("express-rate-limit");
const middlewares = require("./middlewares");
const controllers = require("./controllers");
const schemas_1 = require("./schemas");
const endpoint_1 = require("./endpoint");
class MonitorPlugin extends lisk_sdk_1.BasePlugin {
    constructor() {
        super(...arguments);
        this.configSchema = schemas_1.configSchema;
        this.endpoint = new endpoint_1.Endpoint();
    }
    get nodeModulePath() {
        return __filename;
    }
    async init(context) {
        await super.init(context);
        this._state = {
            forks: {
                forkEventCount: 0,
                blockHeaders: {},
            },
            transactions: {},
            blocks: {},
        };
        this.endpoint.init(this._state, this.apiClient);
    }
    async load() {
        this._app = express();
        this._registerMiddlewares(this.config);
        this._registerControllers();
        this._registerAfterMiddlewares(this.config);
        this._subscribeToEvents();
        this._server = this._app.listen(this.config.port, this.config.host);
    }
    async unload() {
        await new Promise((resolve, reject) => {
            this._server.close(err => {
                if (err) {
                    reject(err);
                    return;
                }
                resolve();
            });
        });
    }
    get state() {
        return this._state;
    }
    _registerMiddlewares(options) {
        this._app.use(cors(options.cors));
        this._app.use(express.json());
        this._app.use((0, express_rate_limit_1.rateLimit)(options.limits));
        this._app.use(middlewares.whiteListMiddleware(options));
    }
    _registerAfterMiddlewares(_options) {
        this._app.use(middlewares.errorMiddleware());
    }
    _registerControllers() {
        this._app.get('/api/prometheus/metrics', controllers.prometheusExport.getData(this.apiClient, this._state));
    }
    _subscribeToEvents() {
        this.apiClient.subscribe('network_newBlock', (data) => {
            const { blockHeader } = data;
            this._handlePostBlock(blockHeader);
        });
        this.apiClient.subscribe('network_newTransaction', (data) => {
            const { transactionIds } = data;
            this._handlePostTransactionAnnounce({ transactionIds });
        });
        this.apiClient.subscribe('chain_forked', (data) => {
            const { blockHeader } = data;
            this._handleFork(blockHeader);
        });
    }
    _handlePostTransactionAnnounce(data) {
        for (const aTransactionId of data.transactionIds) {
            if (this._state.transactions[aTransactionId]) {
                this._state.transactions[aTransactionId].count += 1;
            }
            else {
                this._state.transactions[aTransactionId] = {
                    count: 1,
                    timeReceived: Date.now(),
                };
                this._cleanUpTransactionStats();
            }
        }
    }
    _cleanUpTransactionStats() {
        const expiryTime = 600000;
        for (const transactionID of Object.keys(this._state.transactions)) {
            if (Date.now() - this._state.transactions[transactionID].timeReceived > expiryTime) {
                delete this._state.transactions[transactionID];
            }
        }
    }
    _handleFork(header) {
        this._state.forks.forkEventCount += 1;
        const blockId = header.id;
        if (this._state.forks.blockHeaders[blockId]) {
            this._state.forks.blockHeaders[blockId].timeReceived = Date.now();
        }
        else {
            this._state.forks.blockHeaders[blockId] = {
                blockHeader: header,
                timeReceived: Date.now(),
            };
        }
    }
    _handlePostBlock(header) {
        if (!this._state.blocks[header.id]) {
            this._state.blocks[header.id] = {
                count: 0,
                height: header.height,
            };
        }
        this._state.blocks[header.id].count += 1;
        for (const id of Object.keys(this._state.blocks)) {
            const blockInfo = this._state.blocks[id];
            if (blockInfo.height < header.height - 300) {
                delete this._state.blocks[id];
            }
        }
    }
}
exports.MonitorPlugin = MonitorPlugin;
//# sourceMappingURL=monitor_plugin.js.map